/*
 * clone2fs.c - copy an entire ext2/ext3 file system.
 * Copyright (C) 2008 Michael Riepe, all rights reserved.
 */

#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <time.h>
#include <assert.h>

#include <unistd.h>
#include <fcntl.h>

#include <ext2fs/ext2fs.h>

static const char *outfile = NULL;
static int overwrite = 0;
static int verbose = 1;
static int zero_fill = 0;

static int
xwrite(int fd, char *buf, size_t len) {
	ssize_t n;

	while (len) {
		n = write(fd, buf, len);
		if (n == -1) {
			if (errno == EINTR || errno == EAGAIN) {
				continue;
			}
			perror("write");
			return 1;
		}
		if (n == 0) {
			fprintf(stderr, "short write - out of disk space?\n");
			return 1;
		}
		buf += n;
		len -= n;
	}
	return 0;
}

static int
skip_bad_blocks(ext2_filsys fs) {
	ext2_badblocks_list bblist = NULL;
	ext2_badblocks_iterate iter;
	errcode_t res;
	blk_t blk;

	res = ext2fs_read_bb_inode(fs, &bblist);
	if (res) {
		com_err("ext2fs_read_bb_inode", res, "%s", fs->device_name);
		return -1;
	}
	res = ext2fs_badblocks_list_iterate_begin(bblist, &iter);
	if (res) {
		com_err("ext2fs_badblocks_list_iterate_begin", res, "%s", fs->device_name);
		ext2fs_badblocks_list_free(bblist);
		return -1;
	}
	while (ext2fs_badblocks_list_iterate(iter, &blk)) {
		if (verbose) {
			fprintf(stderr, "Skipping bad block %lu\n", (unsigned long)blk);
		}
		ext2fs_unmark_block_bitmap(fs->block_map, blk);
	}
	ext2fs_badblocks_list_iterate_end(iter);
	ext2fs_badblocks_list_free(bblist);
	return 0;
}

static int
sparse_clone(ext2_filsys fs, int outfd) {
	char *buffer = NULL;
	errcode_t res;
	blk_t nblocks;
	blk_t nused;
	blk_t nblk;
	blk_t i;
	int can_seek;
	unsigned lastperm;
	unsigned permille;
	time_t start;

	res = ext2fs_read_block_bitmap(fs);
	if (res) {
		com_err("ext2fs_read_block_bitmap", res, "%s", fs->device_name);
		return -1;
	}
	if (skip_bad_blocks(fs)) {
		return -1;
	}
	buffer = malloc(fs->blocksize);
	if (buffer == NULL) {
		perror("malloc");
		return -1;
	}
	can_seek = lseek(outfd, (off_t)1024, SEEK_SET) == (off_t)1024;
	lseek(outfd, (off_t)0, SEEK_SET);
	nblocks = (blk_t)fs->super->s_blocks_count;
	nused = 0;
	lastperm = permille = 0;
	time(&start);
	for (nblk = i = 0; i < nblocks; ++i) {
		if (i == 0 || ext2fs_test_block_bitmap(fs->block_map, i)) {
			res = io_channel_read_blk(fs->io, i, 1, buffer);
			if (res) {
				com_err("io_channel_read_blk", res, "%s", fs->device_name);
				free(buffer);
				return -1;
			}
			if (nblk != i) {
				assert(nblk < i);
				assert(can_seek);
				if (lseek(outfd, (off_t)i * fs->blocksize, SEEK_SET) == -1) {
					perror("lseek");
					free(buffer);
					return -1;
				}
			}
			if (xwrite(outfd, buffer, (size_t)fs->blocksize)) {
				free(buffer);
				return -1;
			}
			nblk = i + 1;
			++nused;
		}
		else if (zero_fill || !can_seek) {
			memset(buffer, 0, (size_t)fs->blocksize);
			if (xwrite(outfd, buffer, (size_t)fs->blocksize)) {
				free(buffer);
				return -1;
			}
			nblk = i + 1;
		}
		if (verbose) {
			permille = (i + 1) * 1000ull / nblocks;
			if (lastperm < permille) {
				unsigned run;

				lastperm = permille;
				run = (unsigned)difftime(time(NULL), start);
				fprintf(stderr, "%u.%1u%% processed; %u:%02u:%02u elapsed\r",
					permille / 10, permille % 10,
					run / 3600, (run / 60) % 60, run % 60);
			}
		}
	}
	(void)ftruncate(outfd, (off_t)nblocks * fs->blocksize);
	if (verbose) {
		fprintf(stderr, "\n%lu/%lu blocks used (%4.1f%%)\n",
			(unsigned long)nused, (unsigned long)nblocks,
			1e2 * (double)nused / (double)nblocks);
	}
	free(buffer);
	return 0;
}

static int
clone_ext2fs(const char *devname, int outfd) {
	int status = EXIT_FAILURE;
	ext2_filsys fs;
	errcode_t res;

	res = ext2fs_open(devname, EXT2_FLAG_FORCE, 0, 0, unix_io_manager, &fs);
	if (res) {
		com_err("ext2fs_open", res, "%s", devname);
	}
	else {
		if (!sparse_clone(fs, outfd)) {
			status = EXIT_SUCCESS;
		}
		ext2fs_free(fs);
	}
	return status;
}

static void
usage(int status) {
	fprintf(stderr,
		"Usage: clone2fs [option...] (device|image)\n"
		"Options:\n"
		"  -h          display this help\n"
		"  -o output   write image to <output>; \"-\" means stdout\n"
		"  -O output   write image to <output> even if that already exists\n"
		"  -q          be less verbose\n"
		"  -V          show program version and exit\n"
		"  -Z          zero-fill unused blocks\n"
		"\n");
	exit(status);
}

static void
parse_options(int argc, char **argv) {
	int show_version = 0;
	int c;

	while ((c = getopt(argc, argv, "ho:O:qVZ")) != -1) {
		switch (c) {
			default: break;
			case '?': usage(EXIT_FAILURE); break;
			case 'h': usage(EXIT_SUCCESS); break;
			case 'O': overwrite = 1; outfile = optarg; break;
			case 'o': overwrite = 0; outfile = optarg; break;
			case 'q': verbose = 0; break;
			case 'V': show_version = 1; break;
			case 'Z': zero_fill = 1; break;
		}
	}
	if (show_version) {
		const char *ver, *date;

		fprintf(stderr, "clone2fs %s\n", VERSION_STRING);
		fprintf(stderr, "Copyright (C) 2008 Michael Riepe, all rights reserved.\n");
		ext2fs_get_library_version(&ver, &date);
		fprintf(stderr, "\tUsing EXT2FS Library version %s, %s\n", ver, date);
		exit(EXIT_SUCCESS);
	}
}

int
main(int argc, char **argv) {
	const char *dev;
	int flags;
	int outfd;
	int res;

	initialize_ext2_error_table();
	parse_options(argc, argv);
	if (optind >= argc) {
		fprintf(stderr, "Missing device argument.\n");
		usage(EXIT_FAILURE);
	}
	if (optind + 1 < argc) {
		fprintf(stderr, "Too many arguments.\n");
		usage(EXIT_FAILURE);
	}
	dev = argv[optind];
	if (outfile == NULL) {
		fprintf(stderr, "You must specify an output file.\n");
		usage(EXIT_FAILURE);
	}
	if (strcmp(outfile, "-") == 0) {
		outfile = "standard output";
		outfd = dup(STDOUT_FILENO);
	}
	else {
		flags = O_RDWR | O_CREAT | (overwrite ? O_TRUNC : O_EXCL);
		outfd = open(outfile, flags, 0600);
	}
	if (outfd == -1) {
		perror(outfile);
		exit(EXIT_FAILURE);
	}
	if (isatty(outfd)) {
		fprintf(stderr, "Refusing to write to a terminal.\n");
		exit(EXIT_FAILURE);
	}
	res = clone_ext2fs(dev, outfd);
	close(outfd);
	exit(res);
}
